/*
	This is our ongoing example of a collection (or container) class.
	
	We will be examining the inner workings of collection classes
	over the next couple of weeks.
	
	In the future, instead of having to write your own version of
	a collection class, you can use one of the pre-existing collection
	classes (e.g., ArrayList) that are part of the Java API.
	
	Modifications:
	
	Wed 11/3: Added an insert method to allow us to put new CDs into
	the list so that the CD's are ordered by artist.
	
	
*/

public class CDCollection {

	public static final int DEFAULT_COLLECTION_SIZE = 100;

	//instance variables
	private CD[] cdArray; // the array
	private int cdCount;  // logical size


	// constructors

	public CDCollection (int capacity) {
		cdArray = new CD[capacity];
		cdCount = 0;
	}

	public CDCollection () {
		this(DEFAULT_COLLECTION_SIZE);
	}

	public class CollectionBoundsException extends Exception {
	
		public CollectionBoundsException () {
			super();
		}
		public CollectionBoundsException (String info) {
			super(info);
		}
	}


	// simple accessors

	public CD get (int index) throws CollectionBoundsException {
		if (index < 0 || index >= cdCount) // error
			throw new CollectionBoundsException (index + " is an illegal index into the collection");
		return cdArray[index];
	}
	public int getSize() {
		return cdCount;
	}

	public String toString() {
	// Returns a string representing the entire contents of the CDCollection
		String result = "    CDCollection[" + "used=" + cdCount + "/length=" + cdArray.length + "\n";
		for (int i = 0; i < cdCount; i++)
		    result += "      cdArray["+i+"]= " + cdArray[i] + "\n";
		return result + " ]";
    }
    
    
	// simple mutator
	public void append (CD anotherCD) {
		cdArray[cdCount] = anotherCD;
		cdCount++;
	}
	
	
	

	public void insert (CD newCD) {
	// This method assumes that the current collection of CDs is ordered by the artist name
	// This method adds the new CD into an appropriate position.
	
	// If there is already a CD in the collection by the same artist as the new CD, we
	// we didn't specify where the CD should be put relative to the other CD's by that artist.
	// However, as coded, the new CD is placed last amongst the CD's by the artist
	
	
		// This version of insert shifts the elements from the tail end of the collection
		// to the right while finding the correct position for the new CD.
		
	
		append(newCD); // this makes certain that the array is big enough for shifting
					   // we'll discuss why in Friday's class
		
		int pos = cdCount-2; // The initial value of pos is the last CD in the array 
							 // prior to appending the new CD
		
		while (pos >= 0 && cdArray[pos].getArtist().compareTo(newCD.getArtist())  > 0 ) {
			// there are two reasons to exit the loop
			//   a) the new CD belongs at the head of the array (here, pos = -1)
			//   b) the CD at pos is "less than" the new CD being added		
		
			cdArray[pos+1] = cdArray[pos]; // shift the object at pos to the right by one
			pos--;
		
		}
		
		// when the loop finishes, either 
		//   a) pos = -1; or 
		//   b) the object at pos is less than the new cd
		
		cdArray[pos+1] = newCD;
	}
	
	
	public void insert2 (CD newCD) {
	// This is an alternative algorithm to insert the CD.
	// Here, though, the process is divided into two stages.
	//  1st): determine the position where the new CD belongs
	//  2nd): shift all elements from that position to the right
	
	// This approach is more work because it means that every element of the
	// array is examined no matter how many elements of the array have to be moved.
	
	// For example, if the array had 1000 elements and the new CD should be third from the
	// last, this algorithm looks at 998 cds in stage 1 and then shifts 2 elements in stage 2.
	// In the version we did in class, the algorithm only shifts the 2 elements and stops
	// when it finds that the third needn't be shifted.
	
	// This version always does at least as much work as the other version and the only 
	// time that they do the same amount of work is when the new CD belongs at position 0.
	
	
		append (newCD);  // this makes certain that the array is big enough for shifting
		
		// find the insert position
		int pos = 0;
		while (cdArray[pos].getArtist().compareTo(newCD.getArtist()) < 0) 
			pos++;
		
		// at this point, pos is the place where newCD should be inserted;
		// however, we have to shift to the right all elements from 
		// pos up to (but not including) the last element of the array
		for (int j = cdCount-2; j >= pos; j--)
			cdArray[j+1] = cdArray[j];
		
		// The above for loop could also have been written as:
		
		//		for (int j = cdCount-1; j > pos; j--)
		//			cdArray[j] = cdArray[j-1];
		
		
		// finally, don't forget to put the new CD in the place that's been opened up
		cdArray[pos] = newCD;
	
	}
	
	

		
	// SEARCHING (accessors)

	// a search method that searches for a CD title

	public int searchForTitle(String theTitle) {	
	// returns the location of the CD with the specified title
	// returns -1 if no CD matches
	
		int look = 0;
		while (look < cdCount) {
			if ( theTitle.equals (cdArray[look].getTitle()))
				return look;
			look++;
		}
		return -1; // because the title wasn't found
	}


	// FILTERING (accessors)


	// a filter that produces all CDs in the collection by an artist

	public CDCollection getAllByArtist(String theArtist) {
	// This method returns a CDCollection object consisting of all
	// CDs in this collection that are by the specified artist
	
		CDCollection newColl = new CDCollection(cdCount);
		
		for (int i = 0; i < cdCount; i++) {
			
			if (cdArray[i].getArtist().equals(theArtist))
				newColl.append(cdArray[i]);
		}
		return newColl;
	}

}